# pylint: disable=W0621
"""Asynchronous Python client for IPP."""
import asyncio

from pyipp import IPP


async def main() -> None:
    """Show example of connecting to your IPP print server."""
    async with IPP("ipps://EPSON761251.local:631/ipp/print") as ipp:
        # Get Printer Info
        printer = await ipp.printer()
        print(printer)

    async with IPP("ipp://hp6830.local:631/ipp/print") as ipp:
        # Get Printer Info
        printer = await ipp.printer()
        print(printer)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""Tests for IPP."""
from __future__ import annotations

import os
from typing import Any

DEFAULT_PRINTER_HOST = "epson761251.local"
DEFAULT_PRINTER_PORT = 631
DEFAULT_PRINTER_PATH = "/ipp/print"
DEFAULT_PRINTER_URI = (
    f"ipp://{DEFAULT_PRINTER_HOST}:{DEFAULT_PRINTER_PORT}{DEFAULT_PRINTER_PATH}"
)

# COMMON (NOT DOCUMENTED)
COMMON_PRINTER_ATTRS: dict[str, Any] = {
    "marker-colors": "",
    "marker-high-levels": "",
    "marker-levels": "",
    "marker-low-levels": "",
    "marker-names": "",
    "marker-types": "",
}

# RFC 2911 - IPP 1.1
RFC2911_PRINTER_ATTRS: dict[str, Any] = {
    "attributes-charset": "",
    "attributes-natural-language": "",
    "charset-configured": "",
    "charset-supported": "",
    "color-supported": "",
    "compression-supported": "",
    "copies-default": "",
    "copies-supported": "",
    "document-format-default": "",
    "document-format-supported": "",
    "finishings-default": "",
    "finishings-supported": "",
    "generated-natural-language-supported": "",
    "ipp-versions-supported": "",
    "media-default": "",
    "media-ready": "",
    "media-supported": "",
    "operations-supported": "",
    "orientation-requested-default": "",
    "orientation-requested-supported": "",
    "output-bin-default": "",
    "output-bin-supported": "",
    "pages-per-minute": "",
    "pages-per-minute-color": "",
    "pdl-override-supported": "",
    "print-quality-default": "",
    "print-quality-supported": "",
    "printer-alert": "",
    "printer-alert-description": "",
    "printer-device-id": "",
    "printer-info": "",
    "printer-is-accepting-jobs": "",
    "printer-location": "",
    "printer-make-and-model": "",  # optional
    "printer-more-info": "",
    "printer-name": "",
    "printer-resolution-default": "",
    "printer-state": "",
    "printer-state-reasons": "",
    "printer-up-time": 0,
    "printer-uri": "",
    "printer-uri-supported": "",
    "queued-job-count": "",
    "request-id": "",
    "requested-attributes": "",
    "requesting-user-name": "",
    "sides-default": "",
    "sides-supported": "",
    "status-code": "",
    "status-message": "",
    "uri-authentication-supported": "",
    "uri-security-supported": "",
    "version-number": "",
}

# RFC 3995 - Notifications
RFC3995_PRINTER_ATTRS: dict[str, Any] = {
    "notify-events-default": "",
    "notify-events-supported": "",
    "notify-lease-duration-default": "",
    "notify-lease-duration-supported": "",
    "notify-max-events-supported": "",
    "notify-pull-method-supported": "",
    "printer-state-change-date-time": "",
    "printer-state-change-time": "",
}

# PWG 5100.1 - Finishings
PWG51001_PRINTER_ATTRS: dict[str, Any] = {
    "finishings-col-default": "",
    "finishings-col-ready": "",
}

# PWG 5100.2 - Output Bins
PWG51002_PRINTER_ATTRS: dict[str, Any] = {
    "output-bin-default": "",
    "output-bin-supported": "",
}

# PWG 5100.3
PWG51003_PRINTER_ATTRS: dict[str, Any] = {
    "job-account-id-default": "",
    "job-account-id-supported": "",
    "job-accounting-user-id-default": "",
    "job-accounting-user-id-supported": "",
    "media-col-default": "",
    "media-col-ready": "",
    "media-col-supported": "",
    "media-type-supported": "",
}

# PWG 5100.4
PWG51004_PRINTER_ATTRS: dict[str, Any] = {
    "pwg-raster-document-resolution-supported": "",
    "pwg-raster-document-sheet-back": "",
    "pwg-raster-document-type-supported": "",
}

# PWG 5100.6 - Page Overrides
PWG51006_PRINTER_ATTRS: dict[str, Any] = {
    "overrides-supported": "",
}

# PWG 5100.7
PWG51007_PRINTER_ATTRS: dict[str, Any] = {
    "print-content-optimize-default": "",
    "print-content-optimize-supported": "",
}

# PWG 5100.9 - Alerts
PWG51009_PRINTER_ATTRS: dict[str, Any] = {
    "printer-alert": "",
    "printer-alert-description": "",
}

# PWG 5100.11 - Extended Options Set 2
PWG510011_PRINTER_ATTRS: dict[str, Any] = {
    "feed-orientation-default": "",
    "feed-orientation-supported": "",
    "job-creation-attributes-supported": "",
    "job-ids-supported": "",
    "job-password-supported": "",
    "job-password-encryption-supported": "",
    "media-col-database": "",
    "which-jobs-supported": "",
}

# PWG 5100.13 - Extended Options Sst 3
PWG510013_PRINTER_ATTRS: dict[str, Any] = {
    "document-password-supported": "",
    "identify-actions-default": "",
    "identify-actions-supported": "",
    "ipp-features-supported": "",
    "job-constraints-supported": "",
    "job-preferred-attributes-supported": "",
    "job-resolvers-supported": "",
    "media-bottom-margin-supported": "",
    "media-col-database.media-source-properties": "",
    "media-col-ready.media-source-properties": "",
    "media-left-margin-supported": "",
    "media-right-margin-supported": "",
    "media-source-supported": "",
    "media-top-margin-supported": "",
    "multiple-operation-timeout-action": "",
    "print-color-mode-default": "",
    "print-color-mode-supported": "",
    "print-rendering-intent-default": "",
    "print-rendering-intent-supported": "",
    "printer-charge-info": "",
    "printer-charge-info-url": "",
    "printer-config-change-date-time": "",
    "printer-config-change-time": "",
    "printer-geo-location": "",
    "printer-get-attributes-supported": "",
    "printer-icc-profiles": "",
    "printer-icons": "",
    "printer-mandatory-job-attributes": "",
    "printer-organization": "",
    "printer-organizational-unit": "",
    "printer-supply": "",
    "printer-supply-description": "",
    "printer-supply-info-uri": "",
    "printer-uuid": "",
    "which-jobs-supported": "",
}

# PWG 5100.14 - IPP Everywhere v1
IPPE10_PRINTER_ATTRS: dict[str, Any] = {
    **RFC2911_PRINTER_ATTRS,  # required
    **RFC3995_PRINTER_ATTRS,
    **PWG51002_PRINTER_ATTRS,
    **PWG51003_PRINTER_ATTRS,
    **PWG51004_PRINTER_ATTRS,
    **PWG51006_PRINTER_ATTRS,
    **PWG51007_PRINTER_ATTRS,
    **PWG51009_PRINTER_ATTRS,
    **PWG510011_PRINTER_ATTRS,
    **PWG510013_PRINTER_ATTRS,
    **COMMON_PRINTER_ATTRS,
}

# PWG 5100.12 - IPP 2.0
IPP20_PRINTER_ATTRS: dict[str, Any] = {
    **RFC2911_PRINTER_ATTRS,  # required
    **RFC3995_PRINTER_ATTRS,  # optional
    **PWG51001_PRINTER_ATTRS,  # required
    **PWG51002_PRINTER_ATTRS,  # required
    **PWG51003_PRINTER_ATTRS,  # optional
    **PWG51004_PRINTER_ATTRS,  # optional
    **PWG51006_PRINTER_ATTRS,  # optional
    **PWG51007_PRINTER_ATTRS,  # optional
    **PWG51009_PRINTER_ATTRS,  # recommended
    **PWG510011_PRINTER_ATTRS,  # optional
    **COMMON_PRINTER_ATTRS,
}


def load_fixture(filename: str) -> str:
    """Load a fixture."""
    path = os.path.join(  # noqa: PTH118
        os.path.dirname(__file__),  # noqa: PTH120
        "fixtures",
        filename,
    )
    with open(path, encoding="utf-8") as fptr:  # noqa: PTH123
        return fptr.read()


def load_fixture_binary(filename: str) -> bytes:
    """Load a binary fixture."""
    path = os.path.join(  # noqa: PTH118
        os.path.dirname(__file__),  # noqa: PTH120
        "fixtures",
        filename,
    )
    with open(path, "rb") as fptr:  # noqa: PTH123
        return fptr.read()



"""Asynchronous Python client for IPP."""
from .exceptions import (
    IPPConnectionError,
    IPPConnectionUpgradeRequired,
    IPPError,
    IPPParseError,
    IPPResponseError,
    IPPVersionNotSupportedError,
)
from .ipp import IPP
from .models import (
    Info,
    Marker,
    Printer,
    State,
    Uri,
)

__all__ = [
    "Info",
    "Marker",
    "Printer",
    "State",
    "Uri",
    "IPP",
    "IPPConnectionError",
    "IPPConnectionUpgradeRequired",
    "IPPError",
    "IPPParseError",
    "IPPResponseError",
    "IPPVersionNotSupportedError",
]



"""Asynchronous Python client for IPP."""
from __future__ import annotations

import asyncio
import sys
from dataclasses import dataclass
from importlib import metadata
from socket import gaierror
from struct import error as structerror
from typing import TYPE_CHECKING, Any

import aiohttp
from deepmerge import always_merger
from yarl import URL

from .const import (
    DEFAULT_CHARSET,
    DEFAULT_CHARSET_LANGUAGE,
    DEFAULT_PRINTER_ATTRIBUTES,
    DEFAULT_PROTO_VERSION,
)
from .enums import IppOperation, IppStatus
from .exceptions import (
    IPPConnectionError,
    IPPConnectionUpgradeRequired,
    IPPError,
    IPPParseError,
    IPPResponseError,
    IPPVersionNotSupportedError,
)
from .models import Printer
from .parser import parse as parse_response
from .serializer import encode_dict

if TYPE_CHECKING:
    from collections.abc import Mapping

if sys.version_info >= (3, 11):
    from asyncio import timeout
else:
    from async_timeout import timeout

VERSION = metadata.version(__package__)

@dataclass
class IPP:
    """Main class for handling connections with IPP servers."""

    host: str
    base_path: str = "/ipp/print"
    password: str | None = None
    port: int = 631
    request_timeout: int = 8
    session: aiohttp.client.ClientSession | None = None
    tls: bool = False
    username: str | None = None
    verify_ssl: bool = False
    user_agent: str | None = None
    ipp_version: tuple[int, int] = DEFAULT_PROTO_VERSION

    _close_session: bool = False
    _printer_uri: str = ""
    _printer: Printer | None = None

    def __post_init__(self) -> None:
        """Initialize connection parameters."""
        if self.host.startswith(("ipp://", "ipps://")):
            self._printer_uri = self.host
            printer_uri = URL(self.host)

            if printer_uri.host is not None:
                self.host = printer_uri.host

            if printer_uri.port is not None:
                self.port = printer_uri.port

            self.tls = printer_uri.scheme == "ipps"  # pylint: disable=W0143
            self.base_path = printer_uri.path
        else:
            self._printer_uri = self._build_printer_uri()

        if self.user_agent is None:
            self.user_agent = f"PythonIPP/{VERSION}"

    async def _request(
        self,
        uri: str = "",
        data: Any | None = None,
        params: Mapping[str, str] | None = None,
    ) -> bytes:
        """Handle a request to an IPP server."""
        scheme = "https" if self.tls else "http"

        method = "POST"
        url = URL.build(
            scheme=scheme,
            host=self.host,
            port=self.port,
            path=self.base_path,
        ).join(URL(uri))

        auth = None
        if self.username and self.password:
            auth = aiohttp.BasicAuth(self.username, self.password)

        headers = {
            "User-Agent": self.user_agent,
            "Content-Type": "application/ipp",
            "Accept": "application/ipp, text/plain, */*",
        }

        if self.session is None:
            self.session = aiohttp.ClientSession()
            self._close_session = True

        if isinstance(data, dict):
            data = encode_dict(data)

        try:
            async with timeout(self.request_timeout):
                response = await self.session.request(
                    method,
                    url,
                    auth=auth,
                    data=data,
                    params=params,
                    headers=headers,
                    ssl=self.verify_ssl,
                )
        except asyncio.TimeoutError as exc:
            raise IPPConnectionError(
                "Timeout occurred while connecting to IPP server.",
            ) from exc
        except (aiohttp.ClientError, gaierror) as exc:
            raise IPPConnectionError(
                "Error occurred while communicating with IPP server.",
            ) from exc

        if response.status == 426:
            raise IPPConnectionUpgradeRequired(
                "Connection upgrade required while communicating with IPP server.",
                {"upgrade": response.headers.get("Upgrade")},
            )

        if (response.status // 100) in [4, 5]:
            content = await response.read()
            response.close()

            raise IPPResponseError(
                f"HTTP {response.status}",  # noqa: EM102
                {
                    "content-type": response.headers.get("Content-Type"),
                    "message": content.decode("utf8"),
                    "status-code": response.status,
                },
            )

        return await response.read()

    def _build_printer_uri(self) -> str:
        scheme = "ipps" if self.tls else "ipp"

        return URL.build(
            scheme=scheme,
            host=self.host,
            port=self.port,
            path=self.base_path,
        ).human_repr()

    def _message(self, operation: IppOperation, msg: dict[str, Any]) -> dict[str, Any]:
        """Build a request message to be sent to the server."""
        base = {
            "version": self.ipp_version,
            "operation": operation,
            "request-id": None,  # will get added by serializer if one isn't given
            "operation-attributes-tag": {  # these are required to be in this order
                "attributes-charset": DEFAULT_CHARSET,
                "attributes-natural-language": DEFAULT_CHARSET_LANGUAGE,
                "printer-uri": self._printer_uri,
                "requesting-user-name": "PythonIPP",
            },
        }

        return always_merger.merge(base, msg)  # type: ignore  # noqa: PGH003

    async def execute(
        self,
        operation: IppOperation,
        message: dict[str, Any],
    ) -> dict[str, Any]:
        """Send a request message to the server."""
        message = self._message(operation, message)
        response = await self._request(data=message)

        try:
            parsed = parse_response(response)
        except (structerror, Exception) as exc:  # disable=broad-except
            raise IPPParseError from exc

        if parsed["status-code"] == IppStatus.ERROR_VERSION_NOT_SUPPORTED:
            raise IPPVersionNotSupportedError("IPP version not supported by server")

        if parsed["status-code"] not in range(0x200):
            raise IPPError(
                "Unexpected printer status code",
                {"status-code": parsed["status-code"]},
            )

        return parsed

    async def raw(self, operation: IppOperation, message: dict[str, Any]) -> bytes:
        """Send a request message to the server and return raw response."""
        message = self._message(operation, message)

        return await self._request(data=message)

    async def close(self) -> None:
        """Close open client session."""
        if self.session and self._close_session:
            await self.session.close()

    async def printer(self) -> Printer:
        """Get printer information from server."""
        response_data = await self.execute(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )

        parsed: dict[str, Any] = next(iter(response_data["printers"] or []), {})

        try:
            if self._printer is None:
                self._printer = Printer.from_dict(parsed)
            else:
                self._printer.update_from_dict(parsed)
        except Exception as exc:
            raise IPPParseError from exc

        return self._printer

    async def __aenter__(self) -> IPP:   # noqa: PYI034
        """Async enter."""
        return self

    async def __aexit__(self, *_exec_info: object) -> None:
        """Async exit."""
        await self.close()



"""Models for IPP."""
# pylint: disable=R0912,R0915
from __future__ import annotations

from dataclasses import asdict, dataclass
from datetime import datetime, timedelta, timezone
from typing import Any

from yarl import URL

from .parser import parse_ieee1284_device_id, parse_make_and_model

PRINTER_STATES = {3: "idle", 4: "printing", 5: "stopped"}


@dataclass
class Info:
    """Object holding information from IPP."""

    name: str
    printer_name: str
    printer_uri_supported: list[str]
    uptime: int
    command_set: str | None = None
    location: str | None = None
    manufacturer: str | None = None
    model: str | None = None
    printer_info: str | None = None
    serial: str | None = None
    uuid: str | None = None
    version: str | None = None
    more_info: str | None = None

    @staticmethod
    def from_dict(data: dict[str, Any]) -> Info:
        """Return Info object from IPP response."""
        cmd = None
        name = "IPP Printer"
        name_parts = []
        serial = None
        _printer_name = printer_name = data.get("printer-name", "")
        make_model = data.get("printer-make-and-model", "")
        device_id = data.get("printer-device-id", "")
        uri_supported = data.get("printer-uri-supported", [])
        uuid = data.get("printer-uuid")

        if not isinstance(uri_supported, list):
            uri_supported = [str(uri_supported)]

        for uri in uri_supported:
            if (URL(uri).path.lstrip("/")) == _printer_name.lstrip("/"):
                _printer_name = ""
                break

        make, model = parse_make_and_model(make_model)
        parsed_device_id = parse_ieee1284_device_id(device_id)

        if parsed_device_id.get("MFG") is not None and len(parsed_device_id["MFG"]) > 0:
            make = parsed_device_id["MFG"]
            name_parts.append(make)

        if parsed_device_id.get("MDL") is not None and len(parsed_device_id["MDL"]) > 0:
            model = parsed_device_id["MDL"]
            name_parts.append(model)

        if parsed_device_id.get("CMD") is not None and len(parsed_device_id["CMD"]) > 0:
            cmd = parsed_device_id["CMD"]

        if parsed_device_id.get("SN") is not None and len(parsed_device_id["SN"]) > 0:
            serial = parsed_device_id["SN"]

        if len(make_model) > 0:
            name = make_model
        elif len(name_parts) == 2:
            name = " ".join(name_parts)
        elif len(_printer_name) > 0:
            name = _printer_name

        return Info(
            command_set=cmd,
            location=data.get("printer-location", ""),
            name=name,
            manufacturer=make,
            model=model,
            printer_name=printer_name,
            printer_info=data.get("printer-info"),
            printer_uri_supported=uri_supported,
            serial=serial,
            uptime=data.get("printer-up-time", 0),
            uuid=uuid[9:] if uuid else None,  # strip urn:uuid: from uuid
            version=data.get("printer-firmware-string-version"),
            more_info=data.get("printer-more-info"),
        )


@dataclass
class Marker:
    """Object holding marker (ink) info from IPP."""

    marker_id: int
    marker_type: str
    name: str
    color: str
    level: int
    low_level: int
    high_level: int


@dataclass
class Uri:
    """Object holding URI info from IPP."""

    uri: str
    authentication: str | None
    security: str | None


@dataclass
class State:
    """Object holding the IPP printer state."""

    printer_state: str
    reasons: str | None
    message: str | None

    @staticmethod
    def from_dict(data: dict[str, Any]) -> State:
        """Return State object from IPP response."""
        state = data.get("printer-state", 0)

        if (reasons := data.get("printer-state-reasons")) == "none":
            reasons = None

        return State(
            printer_state=PRINTER_STATES.get(state, state),
            reasons=reasons,
            message=data.get("printer-state-message"),
        )


@dataclass
class Printer:
    """Object holding the IPP printer information."""

    info: Info
    markers: list[Marker]
    state: State
    uris: list[Uri]
    booted_at: datetime

    def as_dict(self) -> dict[str, Any]:
        """Return dictionary version of this printer."""
        return {
            "info": asdict(self.info),
            "state": asdict(self.state),
            "markers": [asdict(marker) for marker in self.markers],
            "uris": [asdict(uri) for uri in self.uris],
            "booted_at": self.booted_at,
        }

    def update_from_dict(self, data: dict[str, Any]) -> Printer:
        """Return updated Printer object from IPP response data."""
        last_uptime = self.info.uptime

        self.info = Info.from_dict(data)
        self.markers = Printer.merge_marker_data(data)
        self.state = State.from_dict(data)
        self.uris = Printer.merge_uri_data(data)

        if self.info.uptime < last_uptime:
            self.booted_at = _utcnow() - timedelta(seconds=self.info.uptime)

        return self


    @staticmethod
    def from_dict(data: dict[str, Any]) -> Printer:
        """Return Printer object from IPP response data."""
        info = Info.from_dict(data)

        return Printer(
            info=info,
            markers=Printer.merge_marker_data(data),
            state=State.from_dict(data),
            uris=Printer.merge_uri_data(data),
            booted_at=(_utcnow() - timedelta(seconds=info.uptime)),
        )

    @staticmethod
    def merge_marker_data(  # noqa: PLR0912, C901
        data: dict[str, Any],
    ) -> list[Marker]:
        """Return Marker data from IPP response."""
        marker_names = []
        marker_colors = []
        marker_levels = []
        marker_types = []
        marker_highs = []
        marker_lows = []

        if not data.get("marker-names"):
            return []

        if isinstance(data["marker-names"], list):
            marker_names = data["marker-names"]
        elif isinstance(data["marker-names"], str):
            marker_names = [data["marker-names"]]

        if not (mlen := len(marker_names)):
            return []

        for _ in range(mlen):
            marker_colors.append("")
            marker_levels.append(-2)
            marker_types.append("unknown")
            marker_highs.append(100)
            marker_lows.append(0)

        if isinstance(data.get("marker-colors"), list):
            for index, list_value in enumerate(data["marker-colors"]):
                if index < mlen:
                    marker_colors[index] = list_value
        elif isinstance(data.get("marker-colors"), str) and mlen == 1:
            marker_colors[0] = data["marker-colors"]

        if isinstance(data.get("marker-levels"), list):
            for index, list_value in enumerate(data["marker-levels"]):
                if index < mlen:
                    marker_levels[index] = list_value
        elif isinstance(data.get("marker-levels"), int) and mlen == 1:
            marker_levels[0] = data["marker-levels"]

        if isinstance(data.get("marker-high-levels"), list):
            for index, list_value in enumerate(data["marker-high-levels"]):
                if index < mlen:
                    marker_highs[index] = list_value
        elif isinstance(data.get("marker-high-levels"), int) and mlen == 1:
            marker_highs[0] = data["marker-high-levels"]

        if isinstance(data.get("marker-low-levels"), list):
            for index, list_value in enumerate(data["marker-low-levels"]):
                if index < mlen:
                    marker_lows[index] = list_value
        elif isinstance(data.get("marker-low-levels"), int) and mlen == 1:
            marker_lows[0] = data["marker-low-levels"]

        if isinstance(data.get("marker-types"), list):
            for index, list_value in enumerate(data["marker-types"]):
                if index < mlen:
                    marker_types[index] = list_value
        elif isinstance(data.get("marker-types"), str) and mlen == 1:
            marker_types[0] = data["marker-types"]

        markers = [
            Marker(
                marker_id=marker_id,
                marker_type=marker_types[marker_id],
                name=marker_names[marker_id],
                color=marker_colors[marker_id],
                level=marker_levels[marker_id],
                high_level=marker_highs[marker_id],
                low_level=marker_lows[marker_id],
            )
            for marker_id in range(mlen)
        ]
        markers.sort(key=lambda x: x.name)

        return markers

    @staticmethod
    def merge_uri_data(data: dict[str, Any]) -> list[Uri]:  # noqa: PLR0912
        """Return URI data from IPP response."""
        _uris: list[str] = []
        auth: list[str | None] = []
        security: list[str | None] = []

        if not data.get("printer-uri-supported"):
            return []

        if isinstance(data["printer-uri-supported"], list):
            _uris = data["printer-uri-supported"]
        elif isinstance(data["printer-uri-supported"], str):
            _uris = [data["printer-uri-supported"]]

        if not (ulen := len(_uris)):
            return []

        for _ in range(ulen):
            auth.append(None)
            security.append(None)

        if isinstance(data.get("uri-authentication-supported"), list):
            for k, list_value in enumerate(data["uri-authentication-supported"]):
                if k < ulen:
                    auth[k] = _str_or_none(list_value)
        elif isinstance(data.get("uri-authentication-supported"), str) and ulen == 1:
            auth[0] = _str_or_none(data["uri-authentication-supported"])

        if isinstance(data.get("uri-security-supported"), list):
            for k, list_value in enumerate(data["uri-security-supported"]):
                if k < ulen:
                    security[k] = _str_or_none(list_value)
        elif isinstance(data.get("uri-security-supported"), str) and ulen == 1:
            security[0] = _str_or_none(data["uri-security-supported"])

        return [
            Uri(
                uri=_uris[uri_id],
                authentication=auth[uri_id],
                security=security[uri_id],
            )
            for uri_id in range(ulen)
        ]


def _utcnow() -> datetime:
    """Return the current date and time in UTC."""
    return datetime.now(tz=timezone.utc)

def _str_or_none(value: str) -> str | None:
    """Return string while handling string representations of None."""
    if value == "none":
        return None

    return value



"""Constants for IPP."""

DEFAULT_CHARSET = "utf-8"
DEFAULT_CHARSET_LANGUAGE = "en-US"

DEFAULT_CLASS_ATTRIBUTES = ["printer-name", "member-names"]

DEFAULT_JOB_ATTRIBUTES = [
    "job-id",
    "job-name",
    "printer-uri",
    "job-state",
    "job-state-reasons",
    "job-hold-until",
    "job-media-progress",
    "job-k-octets",
    "number-of-documents",
    "copies",
    "job-originating-user-name",
]

DEFAULT_PRINTER_ATTRIBUTES = [
    "printer-device-id",
    "printer-name",
    "printer-type",
    "printer-location",
    "printer-info",
    "printer-make-and-model",
    "printer-state",
    "printer-state-message",
    "printer-state-reason",
    "printer-supply",
    "printer-up-time",
    "printer-uri-supported",
    "device-uri",
    "printer-is-shared",
    "printer-more-info",
    "printer-firmware-string-version",
    "marker-colors",
    "marker-high-levels",
    "marker-levels",
    "marker-low-levels",
    "marker-names",
    "marker-types",
]

DEFAULT_PORT = 631
DEFAULT_PROTO_VERSION = (2, 0)



"""Tests for IPP public interface."""
import pytest
from aiohttp import ClientSession
from aresponses import ResponsesMockServer

from pyipp import IPP, Printer
from pyipp.const import DEFAULT_PRINTER_ATTRIBUTES
from pyipp.enums import IppOperation

from . import (
    DEFAULT_PRINTER_HOST,
    DEFAULT_PRINTER_PATH,
    DEFAULT_PRINTER_PORT,
    DEFAULT_PRINTER_URI,
    load_fixture_binary,
)

MATCH_DEFAULT_HOST = f"{DEFAULT_PRINTER_HOST}:{DEFAULT_PRINTER_PORT}"


@pytest.mark.asyncio
async def test_printer(aresponses: ResponsesMockServer) -> None:
    """Test getting IPP printer information."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        printer = await ipp.printer()

        assert printer
        assert isinstance(printer, Printer)


@pytest.mark.asyncio
async def test_printer_update_logic(aresponses: ResponsesMockServer) -> None:
    """Test getting updated IPP printer information."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
        repeat=2,
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        printer = await ipp.printer()

        assert printer
        assert isinstance(printer, Printer)

        printer = await ipp.printer()
        assert printer
        assert isinstance(printer, Printer)


@pytest.mark.asyncio
async def test_raw(aresponses: ResponsesMockServer) -> None:
    """Test raw method is handled correctly."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        response = await ipp.raw(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )

        assert response
        assert isinstance(response, bytes)



# pylint: disable=W0621
"""Asynchronous Python client for IPP."""
import asyncio

from pyipp import IPP
from pyipp.enums import IppOperation


async def main() -> None:
    """Show example of printing via IPP print server."""
    pdf_file = "/path/to/pdf.pfd"
    with open(pdf_file, "rb") as f:  # noqa: PTH123, ASYNC230
        content = f.read()

    async with IPP("ipp://192.168.1.92:631/ipp/print") as ipp:
        response = await ipp.execute(
            IppOperation.PRINT_JOB,
            {
                "operation-attributes-tag": {
                    "requesting-user-name": "Me",
                    "job-name": "My Test Job",
                    "document-format": "application/pdf",
                },
                "data": content,
            },
        )

        print(response)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""Response Parser for IPP."""
from __future__ import annotations

import logging
import struct
from datetime import datetime, timedelta, timezone
from typing import Any

from .enums import ATTRIBUTE_ENUM_MAP, IppTag
from .exceptions import IPPParseError

_LOGGER = logging.getLogger(__name__)


def parse_ieee1284_device_id(device_id: str) -> dict[str, str]:
    """Parse IEEE 1284 device id for common device info."""
    if not device_id:
        return {}

    device_id = device_id.strip(";")
    device_info: dict[str, str] = {}

    for pair in device_id.split(";"):
        key, value = pair.split(":", 2)
        device_info[key.strip()] = value.strip()

    if not device_info.get("MANUFACTURER") and device_info.get("MFG"):
        device_info["MANUFACTURER"] = device_info["MFG"]

    if not device_info.get("MODEL") and device_info.get("MDL"):
        device_info["MODEL"] = device_info["MDL"]

    if not device_info.get("COMMAND SET") and device_info.get("CMD"):
        device_info["COMMAND SET"] = device_info["CMD"]

    return device_info


def parse_collection(data: bytes, offset: int) -> tuple[dict[str, Any], int]:
    """Parse member attributes from IPP collection."""
    collection_data: dict[str, Any] = {}
    member_name: str = ""

    _LOGGER.debug("#BeginCollectionOffset: %s", offset)

    while struct.unpack_from("b", data, offset)[0] != IppTag.END_COLLECTION.value:
        attribute, next_attr_offset = parse_attribute(data, offset, member_name)

        if attribute["tag"] == IppTag.MEMBER_NAME.value:
            member_name = attribute["value"]
        elif member_name:
            collection_data[member_name] = attribute["value"]

        offset = next_attr_offset

    _LOGGER.debug("#EndCollectionOffset: %s", offset)

    # skip over end of collection marker
    _, next_attr_offset = parse_attribute(data, offset)

    return collection_data, next_attr_offset


# pylint: disable=R0912,R0915
def parse_attribute(  # noqa: PLR0912, PLR0915
    data: bytes,
    offset: int,
    prev_attr_name: str = "",
) -> tuple[dict[str, Any], int]:
    """Parse attribute from IPP data.

    1 byte: Tag - b
    2 byte: Name Length - h
    N bytes: Name - direct access
    2 byte: Value Length -h
    N bytes: Value - direct access
    """
    _LOGGER.debug("Parsing Attribute at offset %s", offset)

    attribute = {"tag": struct.unpack_from(">b", data, offset)[0]}
    offset += 1

    attribute["name-length"] = struct.unpack_from(">h", data, offset)[0]
    offset += 2

    offset_length = offset + attribute["name-length"]
    attribute["name"] = data[offset:offset_length].decode("utf-8")
    offset += attribute["name-length"]

    resolved_attr_name = attribute["name"] or prev_attr_name

    if attribute["name"]:
        _LOGGER.debug(
            "Attribute Name: %s (%s)", attribute["name"], hex(attribute["tag"]),
        )
    else:
        _LOGGER.debug("Attribute Tag: %s", hex(attribute["tag"]))

    attribute["value-length"] = struct.unpack_from(">h", data, offset)[0]
    offset += 2

    _LOGGER.debug("Attribute Value Offset: %s", offset)
    _LOGGER.debug("Attribute Value Length: %s", attribute["value-length"])

    if attribute["tag"] in (IppTag.ENUM.value, IppTag.INTEGER.value):
        attribute["value"] = struct.unpack_from(">i", data, offset)[0]

        if (
            attribute["tag"] == IppTag.ENUM.value
            and resolved_attr_name in ATTRIBUTE_ENUM_MAP
        ):
            enum_class = ATTRIBUTE_ENUM_MAP[resolved_attr_name]
            attribute["value"] = enum_class(attribute["value"])

        offset += 4
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.BOOLEAN.value:
        attribute["value"] = struct.unpack_from(">?", data, offset)[0]
        offset += 1
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.DATE.value:
        if attribute["value-length"] != 11:
            raise IPPParseError(
                f'Invalid DATE size {attribute["value-length"]}',  # noqa: EM102
            )

        raw_date = dict(
            zip(
                (
                    "year",
                    "month",
                    "day",
                    "hour",
                    "minute",
                    "second",
                    "decisecond",
                    "tz_dir",
                    "tz_hour",
                    "tz_minute",
                ),
                struct.unpack_from(">hbbbbbbcbb", data, offset),
            ),
        )
        raw_date["microsecond"] = raw_date.pop("decisecond") * 100_000
        raw_date["tzinfo"] = timezone(
            {b"+": 1, b"-": -1}[raw_date.pop("tz_dir")]
            * timedelta(
                hours=raw_date.pop("tz_hour"),
                minutes=raw_date.pop("tz_minute"),
            ),
        )

        attribute["value"] = datetime(**raw_date)  # noqa: DTZ001
        offset += attribute["value-length"]
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.RESERVED_STRING.value:
        if attribute["value-length"] > 0:
            offset_length = offset + attribute["value-length"]
            attribute["value"] = data[offset:offset_length].decode("utf-8")
            offset += attribute["value-length"]
        else:
            attribute["value"] = None

        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.RANGE.value:
        attribute["value"] = []
        for i in range(int(attribute["value-length"] / 4)):
            attribute["value"].append(struct.unpack_from(">i", data, offset + i * 4)[0])
        offset += attribute["value-length"]
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.RESOLUTION.value:
        attribute["value"] = struct.unpack_from(">iib", data, offset)
        offset += attribute["value-length"]
    elif attribute["tag"] in (IppTag.TEXT_LANG.value, IppTag.NAME_LANG.value):
        attribute["language-length"] = struct.unpack_from(">h", data, offset)[0]
        offset += 2

        offset_length = offset + attribute["language-length"]
        attribute["language"] = data[offset:offset_length].decode("utf-8")
        offset += attribute["language-length"]
        _LOGGER.debug("Attribute Language: %s", attribute["language"])

        attribute["text-length"] = struct.unpack_from(">h", data, offset)[0]
        offset += 2
        _LOGGER.debug("Attribute Text Length: %s", attribute["text-length"])

        offset_length = offset + attribute["text-length"]
        attribute["value"] = data[offset:offset_length].decode("utf-8")
        offset += attribute["text-length"]
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.BEGIN_COLLECTION.value:
        offset += attribute["value-length"]
        collection, new_offset = parse_collection(data, offset)
        attribute["value"] = collection
        offset = new_offset
        _LOGGER.debug("Attribute Value: %s", attribute["value"])
    elif attribute["tag"] == IppTag.END_COLLECTION.value:
        offset += attribute["value-length"]
    else:
        offset_length = offset + attribute["value-length"]
        attribute["value"] = data[offset:offset_length]
        _LOGGER.debug("Attribute Bytes: %s", attribute["value"])

        attribute["value"] = attribute["value"].decode("utf-8", "ignore")
        offset += attribute["value-length"]
        _LOGGER.debug("Attribute Value: %s", attribute["value"])

    return attribute, offset


# pylint: disable=R0912,R0915
def parse(  # noqa: PLR0912, PLR0915
    raw_data: bytes,
    contains_data: bool = False,  # noqa: FBT001, FBT002
) -> dict[str, Any]:
    r"""Parse raw IPP data.

    1 byte: Protocol Major Version - b
    1 byte: Protocol Minor Version - b
    2 byte: Operation ID/Status Code - h
    4 byte: Request ID - i

    1 byte: Operation Attribute Byte (\0x01)

    N Mal: Attributes

    1 byte: Attribute End Byte (\0x03)
    """
    data: dict[str, Any] = {}
    offset = 0

    _LOGGER.debug("Parsing IPP Data")

    data["version"] = struct.unpack_from(">bb", raw_data, offset)
    offset += 2

    _LOGGER.debug("IPP Version: %s", data["version"])

    data["status-code"] = struct.unpack_from(">h", raw_data, offset)[0]
    offset += 2

    _LOGGER.debug("IPP Status Code: %s", data["status-code"])

    data["request-id"] = struct.unpack_from(">i", raw_data, offset)[0]
    offset += 4

    data["operation-attributes"] = []
    data["unsupported-attributes"] = []
    data["jobs"] = []
    data["printers"] = []
    data["data"] = b""

    attribute_key = ""
    previous_attribute_name = ""
    tmp_data: dict[str, Any] = {}

    while struct.unpack_from("b", raw_data, offset)[0] != IppTag.END.value:
        # check for operation, job or printer attribute start byte
        # if tmp data and attribute key is set, another operation was sent
        # add it and reset tmp data
        if struct.unpack_from("b", raw_data, offset)[0] == IppTag.OPERATION.value:
            if tmp_data and attribute_key:
                data[attribute_key].append(tmp_data)
                tmp_data = {}

            attribute_key = "operation-attributes"
            offset += 1
        elif struct.unpack_from("b", raw_data, offset)[0] == IppTag.JOB.value:
            if tmp_data and attribute_key:
                data[attribute_key].append(tmp_data)
                tmp_data = {}

            attribute_key = "jobs"
            offset += 1
        elif struct.unpack_from("b", raw_data, offset)[0] == IppTag.PRINTER.value:
            if tmp_data and attribute_key:
                data[attribute_key].append(tmp_data)
                tmp_data = {}

            attribute_key = "printers"
            offset += 1
        elif (
            struct.unpack_from("b", raw_data, offset)[0]
            == IppTag.UNSUPPORTED_GROUP.value
        ):
            if tmp_data and attribute_key:
                data[attribute_key].append(tmp_data)
                tmp_data = {}

            attribute_key = "unsupported-attributes"
            offset += 1

        attribute, new_offset = parse_attribute(
            raw_data, offset, previous_attribute_name,
        )

        # if attribute has a name -> add it
        # if attribute doesn't have a name -> it is part of an array
        if attribute["name"]:
            tmp_data[attribute["name"]] = attribute["value"]
            previous_attribute_name = attribute["name"]
        elif previous_attribute_name:
            # check if attribute is already an array
            # else convert it to an array
            if isinstance(tmp_data[previous_attribute_name], list):
                tmp_data[previous_attribute_name].append(attribute["value"])
            else:
                tmp_value = tmp_data[previous_attribute_name]
                tmp_data[previous_attribute_name] = [tmp_value, attribute["value"]]

        offset = new_offset

    if isinstance(data[attribute_key], list):
        data[attribute_key].append(tmp_data)

    if isinstance(data["operation-attributes"], list):
        data["operation-attributes"] = data["operation-attributes"][0]

    if contains_data:
        offset_start = offset + 1
        data["data"] = raw_data[offset_start:]

    return data


def parse_make_and_model(make_and_model: str) -> tuple[str, str]:
    """Parse make and model for separate device make and model."""
    if not (make_and_model := make_and_model.strip()):
        return ("Unknown", "Unknown")

    make = "Unknown"
    model = "Unknown"
    found_make = False
    known_makes = [
        "brother",
        "canon",
        "epson",
        "kyocera",
        "hp",
        "xerox",
    ]

    test_against = make_and_model.lower()
    for known_make in known_makes:
        if test_against.startswith(known_make):
            found_make = True
            mlen = len(known_make)
            make = make_and_model[:mlen]
            model = make_and_model[mlen:].strip()
            break

    if not found_make:
        split = make_and_model.split(None, 1)
        make = split[0]

        if len(split) == 2:
            model = split[1].strip()

    return (make, model)



"""Tests for Serializer."""
from pyipp import serializer
from pyipp.const import DEFAULT_CHARSET, DEFAULT_CHARSET_LANGUAGE, DEFAULT_PROTO_VERSION
from pyipp.enums import IppOperation, IppTag

from . import load_fixture_binary


def test_construct_attribute_values() -> None:
    """Test the construct_attribute_values method."""
    result = serializer.construct_attribute_values(
        IppTag.INTEGER,
        IppOperation.GET_PRINTER_ATTRIBUTES,
    )
    assert result == b"\x00\x04\x00\x00\x00\x0b"

    result = serializer.construct_attribute_values(
        IppTag.ENUM,
        IppOperation.GET_PRINTER_ATTRIBUTES,
    )
    assert result == b"\x00\x04\x00\x00\x00\x0b"

    result = serializer.construct_attribute_values(
        IppTag.BOOLEAN,
        "0",
    )
    assert result == b"\x00\x01\x01"

    result = serializer.construct_attribute_values(
        IppTag.URI,
        "ipps://localhost:631",
    )
    assert result == b"\x00\x14ipps://localhost:631"


def test_construct_attribute() -> None:
    """Test the construct_attribute method."""
    result = serializer.construct_attribute("attributes-charset", DEFAULT_CHARSET)
    assert result == b"G\x00\x12attributes-charset\x00\x05utf-8"

    result = serializer.construct_attribute(
        "operations-supported",
        [IppOperation.GET_PRINTER_ATTRIBUTES],
    )
    assert result == b"#\x00\x14operations-supported\x00\x04\x00\x00\x00\x0b"


def test_construct_attribute_no_tag_unmapped() -> None:
    """Test the construct_attribute method with no tag and unmapped attribute name."""
    result = serializer.construct_attribute(
        "no-tag-unmapped",
        None,
    )

    assert result == b""


def test_encode_dict() -> None:
    """Test the encode_dict method."""
    result = serializer.encode_dict(
        {
            "version": DEFAULT_PROTO_VERSION,
            "operation": IppOperation.GET_PRINTER_ATTRIBUTES,
            "request-id": 1,
            "operation-attributes-tag": {
                "attributes-charset": DEFAULT_CHARSET,
                "attributes-natural-language": DEFAULT_CHARSET_LANGUAGE,
                "printer-uri": "ipp://printer.example.com:361/ipp/print",
                "requesting-user-name": "PythonIPP",
            },
        },
    )

    assert result == load_fixture_binary(
        "serializer/get-printer-attributes-request-000.bin",
    )



"""Exceptions for IPP."""


class IPPError(Exception):
    """Generic IPP exception."""


class IPPConnectionError(IPPError):
    """IPP connection exception."""


class IPPConnectionUpgradeRequired(IPPError):  # noqa: N818
    """IPP connection upgrade requested."""


class IPPParseError(IPPError):
    """IPP parse exception."""


class IPPResponseError(IPPError):
    """IPP response exception."""


class IPPVersionNotSupportedError(IPPError):
    """IPP version not supported."""



"""Tests for IPP Client."""
import asyncio

import pytest
from aiohttp import ClientResponse, ClientSession
from aresponses import Response, ResponsesMockServer

from pyipp import IPP
from pyipp.const import DEFAULT_PRINTER_ATTRIBUTES
from pyipp.enums import IppOperation
from pyipp.exceptions import (
    IPPConnectionError,
    IPPConnectionUpgradeRequired,
    IPPError,
    IPPParseError,
    IPPVersionNotSupportedError,
)

from . import (
    DEFAULT_PRINTER_HOST,
    DEFAULT_PRINTER_PATH,
    DEFAULT_PRINTER_PORT,
    DEFAULT_PRINTER_URI,
    load_fixture_binary,
)

MATCH_DEFAULT_HOST = f"{DEFAULT_PRINTER_HOST}:{DEFAULT_PRINTER_PORT}"
NON_STANDARD_PORT = 3333


@pytest.mark.asyncio
async def test_ipp_request(aresponses: ResponsesMockServer) -> None:
    """Test IPP response is handled correctly."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        response = await ipp.execute(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )
        assert response["status-code"] == 0


@pytest.mark.asyncio
async def test_internal_session(aresponses: ResponsesMockServer) -> None:
    """Test IPP response is handled correctly."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with IPP(DEFAULT_PRINTER_URI) as ipp:
        response = await ipp.execute(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )
        assert response["status-code"] == 0


@pytest.mark.asyncio
async def test_request_port(aresponses: ResponsesMockServer) -> None:
    """Test the IPP server running on non-standard port."""
    aresponses.add(
        f"{DEFAULT_PRINTER_HOST}:{NON_STANDARD_PORT}",
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(
            host=DEFAULT_PRINTER_HOST,
            port=NON_STANDARD_PORT,
            base_path=DEFAULT_PRINTER_PATH,
            session=session,
        )
        response = await ipp.execute(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )
        assert response["status-code"] == 0


@pytest.mark.asyncio
async def test_request_tls(aresponses: ResponsesMockServer) -> None:
    """Test the IPP server over TLS."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-epsonxp6000.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(
            host=DEFAULT_PRINTER_HOST,
            port=DEFAULT_PRINTER_PORT,
            tls=True,
            base_path=DEFAULT_PRINTER_PATH,
            session=session,
        )
        response = await ipp.execute(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "operation-attributes-tag": {
                    "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                },
            },
        )
        assert response["status-code"] == 0


@pytest.mark.asyncio
async def test_timeout(aresponses: ResponsesMockServer) -> None:
    """Test request timeout from the IPP server."""

    # Faking a timeout by sleeping
    async def response_handler(_: ClientResponse) -> Response:
        await asyncio.sleep(2)
        return Response(body="Timeout!")

    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        response_handler,
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session, request_timeout=1)
        with pytest.raises(IPPConnectionError):
            assert await ipp.execute(
                IppOperation.GET_PRINTER_ATTRIBUTES,
                {
                    "operation-attributes-tag": {
                        "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                    },
                },
            )


@pytest.mark.asyncio
async def test_http_error404(aresponses: ResponsesMockServer) -> None:
    """Test HTTP 404 response handling."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(text="Not Found!", status=404),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        with pytest.raises(IPPError):
            assert await ipp.execute(
                IppOperation.GET_PRINTER_ATTRIBUTES,
                {
                    "operation-attributes-tag": {
                        "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                    },
                },
            )


@pytest.mark.asyncio
async def test_http_error426(aresponses: ResponsesMockServer) -> None:
    """Test HTTP 426 response handling."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            text="Upgrade Required",
            headers={"Upgrade": "TLS/1.0, HTTP/1.1"},
            status=426,
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        with pytest.raises(IPPConnectionUpgradeRequired):
            assert await ipp.execute(
                IppOperation.GET_PRINTER_ATTRIBUTES,
                {
                    "operation-attributes-tag": {
                        "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                    },
                },
            )


@pytest.mark.asyncio
async def test_unexpected_response(aresponses: ResponsesMockServer) -> None:
    """Test unexpected response handling."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(text="Surprise!", status=200),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        with pytest.raises(IPPParseError):
            assert await ipp.execute(
                IppOperation.GET_PRINTER_ATTRIBUTES,
                {
                    "operation-attributes-tag": {
                        "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                    },
                },
            )


@pytest.mark.asyncio
async def test_ipp_error_0x0503(aresponses: ResponsesMockServer) -> None:
    """Test IPP Error 0x0503 response handling."""
    aresponses.add(
        MATCH_DEFAULT_HOST,
        DEFAULT_PRINTER_PATH,
        "POST",
        aresponses.Response(
            status=200,
            headers={"Content-Type": "application/ipp"},
            body=load_fixture_binary("get-printer-attributes-error-0x0503.bin"),
        ),
    )

    async with ClientSession() as session:
        ipp = IPP(DEFAULT_PRINTER_URI, session=session)
        with pytest.raises(IPPVersionNotSupportedError):
            assert await ipp.execute(
                IppOperation.GET_PRINTER_ATTRIBUTES,
                {
                    "operation-attributes-tag": {
                        "requested-attributes": DEFAULT_PRINTER_ATTRIBUTES,
                    },
                },
            )



"""Tests for Parser."""
import pytest
from syrupy.assertion import SnapshotAssertion

from pyipp import IPPParseError, parser
from pyipp.const import DEFAULT_CHARSET, DEFAULT_CHARSET_LANGUAGE, DEFAULT_PROTO_VERSION
from pyipp.enums import IppOperation

from . import load_fixture_binary

RESPONSE_GET_PRINTER_ATTRIBUTES = load_fixture_binary(
    "get-printer-attributes-response-000.bin",
)

MOCK_IEEE1284_DEVICE_ID = "MFG:EPSON;CMD:ESCPL2,BDC,D4,D4PX,ESCPR7,END4,GENEP,URF;MDL:XP-6000 Series;CLS:PRINTER;DES:EPSON XP-6000 Series;CID:EpsonRGB;FID:FXN,DPA,WFA,ETN,AFN,DAN,WRA;RID:20;DDS:022500;ELG:1000;SN:583434593035343012;URF:CP1,PQ4-5,OB9,OFU0,RS360,SRGB24,W8,DM3,IS1-7-6,V1.4,MT1-3-7-8-10-11-12;"


def test_parse() -> None:
    """Test the parse method."""
    result = parser.parse(RESPONSE_GET_PRINTER_ATTRIBUTES)
    assert result == {
        "data": b"",
        "jobs": [],
        "operation-attributes": {
            "attributes-charset": DEFAULT_CHARSET,
            "attributes-natural-language": DEFAULT_CHARSET_LANGUAGE,
            "printer-uri": "ipp://printer.example.com:361/ipp/print",
            "requesting-user-name": "PythonIPP",
        },
        "printers": [],
        "request-id": 1,
        "status-code": IppOperation.GET_PRINTER_ATTRIBUTES,
        "unsupported-attributes": [],
        "version": DEFAULT_PROTO_VERSION,
    }


def test_parse_attribute() -> None:
    """Test the parse_attribute method."""
    result = parser.parse_attribute(RESPONSE_GET_PRINTER_ATTRIBUTES, 9)
    assert result == (
        {
            "name": "attributes-charset",
            "name-length": 18,
            "tag": 71,
            "value": "utf-8",
            "value-length": 5,
        },
        37,
    )


def test_parse_attribute_reserved_string() -> None:
    """Test the parse_attribute method when provided a reserved string."""
    result = parser.parse_attribute(b"C\x00\x0freserved-string\x00\x04yoda", 0)
    assert result == (
        {
            "name": "reserved-string",
            "name-length": 15,
            "tag": 67,
            "value": "yoda",
            "value-length": 4,
        },
        24,
    )

    result = parser.parse_attribute(b"C\x00\x0freserved-string\x00\x00", 0)
    assert result == (
        {
            "name": "reserved-string",
            "name-length": 15,
            "tag": 67,
            "value": None,
            "value-length": 0,
        },
        20,
    )


def test_parse_attribute_invalid_date() -> None:
    """Test the parse_attribute method when provided an invalid date."""
    invalid = b"1\x00\x14printer-current-time\x00\x0299"

    with pytest.raises(IPPParseError):
        parser.parse_attribute(invalid, 0)


def test_parse_ieee1284_device_id() -> None:
    """Test the parse_ieee1284_device_id method."""
    result = parser.parse_ieee1284_device_id(MOCK_IEEE1284_DEVICE_ID)

    assert result
    assert result["MFG"] == "EPSON"
    assert result["MDL"] == "XP-6000 Series"
    assert result["SN"] == "583434593035343012"
    assert result["CMD"] == "ESCPL2,BDC,D4,D4PX,ESCPR7,END4,GENEP,URF"

    assert result["MANUFACTURER"] == result["MFG"]
    assert result["MODEL"] == result["MDL"]
    assert result["COMMAND SET"] == result["CMD"]


def test_parse_ieee1284_device_id_manufacturer_only() -> None:
    """Test the parse_ieee1284_device_id method with only a manufacturer."""
    result = parser.parse_ieee1284_device_id("MANUFACTURER:EPSON")

    assert result == {
        "MANUFACTURER": "EPSON",
    }


def test_parse_ieee1284_device_id_empty() -> None:
    """Test the parse_ieee1284_device_id method with empty string."""
    result = parser.parse_ieee1284_device_id("")

    assert isinstance(result, dict)


def test_parse_make_and_model() -> None:
    """Test the parse_make_and_model method."""
    result = parser.parse_make_and_model("")
    assert result == ("Unknown", "Unknown")

    # generic fallback for unknown brands
    result = parser.parse_make_and_model("IPP")
    assert result == ("IPP", "Unknown")

    result = parser.parse_make_and_model("IPP Printer")
    assert result == ("IPP", "Printer")

    # known brands
    result = parser.parse_make_and_model("EPSON XP-6000 Series")
    assert result == ("EPSON", "XP-6000 Series")

    result = parser.parse_make_and_model("HP Officejet Pro 6830")
    assert result == ("HP", "Officejet Pro 6830")

    result = parser.parse_make_and_model("HP Photosmart D110 Series")
    assert result == ("HP", "Photosmart D110 Series")


def test_parse_brother_mfcj5320dw(snapshot: SnapshotAssertion) -> None:
    """Test the parse method against response from Brother MFC-J5320DW."""
    response = load_fixture_binary("get-printer-attributes-brother-mfcj5320dw.bin")

    result = parser.parse(response)
    assert result == snapshot


def test_parse_epson_xp6000(snapshot: SnapshotAssertion) -> None:
    """Test the parse method against response from Epson XP-6000 Series."""
    response = load_fixture_binary("get-printer-attributes-epsonxp6000.bin")

    result = parser.parse(response)
    assert result == snapshot


def test_parse_kyocera_ecosys_m2540dn(snapshot: SnapshotAssertion) -> None:
    """Test the parse method against response from Kyocera Ecosys M2540DN."""
    response = load_fixture_binary(
        "get-printer-attributes-kyocera-ecosys-m2540dn-001.bin",
    )

    result = parser.parse(response)
    assert result == snapshot



"""Data Serializer for IPP."""
from __future__ import annotations

import logging
import random
import struct
from typing import Any

from .const import DEFAULT_PROTO_VERSION
from .enums import IppTag
from .tags import ATTRIBUTE_TAG_MAP

_LOGGER = logging.getLogger(__name__)


def construct_attribute_values(tag: IppTag, value: Any) -> bytes:
    """Serialize the attribute values into IPP format."""
    byte_str = b""

    if tag in (IppTag.INTEGER, IppTag.ENUM):
        byte_str += struct.pack(">h", 4)
        byte_str += struct.pack(">i", value)
    elif tag == IppTag.BOOLEAN:
        byte_str += struct.pack(">h", 1)
        byte_str += struct.pack(">?", value)
    else:
        encoded_value = value.encode("utf-8")
        byte_str += struct.pack(">h", len(encoded_value))
        byte_str += encoded_value

    return byte_str


def construct_attribute(name: str, value: Any, tag: IppTag | None = None) -> bytes:
    """Serialize the attribute into IPP format."""
    byte_str = b""

    if not tag and not (tag := ATTRIBUTE_TAG_MAP.get(name, None)):
        _LOGGER.debug("Unknown IppTag for %s", name)
        return byte_str

    if isinstance(value, (list, tuple, set)):
        for index, list_value in enumerate(value):
            byte_str += struct.pack(">b", tag.value)

            if index == 0:
                byte_str += struct.pack(">h", len(name))
                byte_str += name.encode("utf-8")
            else:
                byte_str += struct.pack(">h", 0)

            byte_str += construct_attribute_values(tag, list_value)
    else:
        byte_str = struct.pack(">b", tag.value)

        byte_str += struct.pack(">h", len(name))
        byte_str += name.encode("utf-8")

        byte_str += construct_attribute_values(tag, value)

    return byte_str


def encode_dict(data: dict[str, Any]) -> bytes:
    """Serialize a dictionary of data into IPP format."""
    version = data["version"] or DEFAULT_PROTO_VERSION
    operation = data["operation"]

    if (request_id := data.get("request-id")) is None:
        request_id = random.choice(range(10000, 99999))  # nosec  # noqa: S311

    encoded = struct.pack(">bb", *version)
    encoded += struct.pack(">h", operation.value)
    encoded += struct.pack(">i", request_id)

    encoded += struct.pack(">b", IppTag.OPERATION.value)

    if isinstance(data.get("operation-attributes-tag"), dict):
        for attr, value in data["operation-attributes-tag"].items():
            encoded += construct_attribute(attr, value)

    if isinstance(data.get("job-attributes-tag"), dict):
        encoded += struct.pack(">b", IppTag.JOB.value)

        for attr, value in data["job-attributes-tag"].items():
            encoded += construct_attribute(attr, value)

    if isinstance(data.get("printer-attributes-tag"), dict):
        encoded += struct.pack(">b", IppTag.PRINTER.value)

        for attr, value in data["printer-attributes-tag"].items():
            encoded += construct_attribute(attr, value)

    encoded += struct.pack(">b", IppTag.END.value)

    if "data" in data:
        encoded += data["data"]

    return encoded



# pylint: disable=W0621
"""Asynchronous Python client for IPP."""
import asyncio

from pyipp import IPP
from pyipp.enums import IppOperation


async def main() -> None:
    """Show example of connecting to your IPP print server."""
    async with IPP("ipps://192.168.1.92:631/ipp/print") as ipp:
        response = await ipp.raw(
            IppOperation.GET_PRINTER_ATTRIBUTES,
            {
                "version": (2, 0),  # try (1, 1) for older devices
                "operation-attributes-tag": {
                    "requested-attributes": [
                        "printer-device-id",
                        "printer-name",
                        "printer-type",
                        "printer-location",
                        "printer-info",
                        "printer-make-and-model",
                        "printer-state",
                        "printer-state-message",
                        "printer-state-reason",
                        "printer-supply",
                        "printer-up-time",
                        "printer-uri-supported",
                        "device-uri",
                        "printer-is-shared",
                        "printer-more-info",
                        "printer-firmware-string-version",
                        "marker-colors",
                        "marker-high-levels",
                        "marker-levels",
                        "marker-low-levels",
                        "marker-names",
                        "marker-types",
                    ],
                },
            },
        )

        with open("printer-attributes.bin", "wb") as file:  # noqa: PTH123, ASYNC230
            file.write(response)
            file.close()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



"""Enumerators for IPP."""
from enum import IntEnum


class IppStatus(IntEnum):
    """Represent the ENUMs of a status response."""

    CUPS_INVALID = -1
    OK = 0x0000
    OK_IGNORED_OR_SUBSTITUTED = 0x0001
    OK_CONFLICTING = 0x0002
    OK_IGNORED_SUBSCRIPTIONS = 0x0003
    OK_IGNORED_NOTIFICATIONS = 0x0004
    OK_TOO_MANY_EVENTS = 0x0005
    OK_BUT_CANCEL_SUBSCRIPTION = 0x0006
    OK_EVENTS_COMPLETE = 0x0007
    REDIRECTION_OTHER_SITE = 0x0200
    CUPS_SEE_OTHER = 0x0280
    ERROR_BAD_REQUEST = 0x0400
    ERROR_FORBIDDEN = 0x0401
    ERROR_NOT_AUTHENTICATED = 0x0402
    ERROR_NOT_AUTHORIZED = 0x0403
    ERROR_NOT_POSSIBLE = 0x0404
    ERROR_TIMEOUT = 0x0405
    ERROR_NOT_FOUND = 0x0406
    ERROR_GONE = 0x0407
    ERROR_REQUEST_ENTITY = 0x0408
    ERROR_REQUEST_VALUE = 0x0409
    ERROR_DOCUMENT_FORMAT_NOT_SUPPORTED = 0x040A
    ERROR_ATTRIBUTES_OR_VALUES = 0x040B
    ERROR_URI_SCHEME = 0x040C
    ERROR_CHARSET = 0x040D
    ERROR_CONFLICTING = 0x040E
    ERROR_COMPRESSION_ERROR = 0x040F
    ERROR_DOCUMENT_FORMAT_ERROR = 0x0410
    ERROR_DOCUMENT_ACCESS = 0x0411
    ERROR_ATTRIBUTES_NOT_SETTABLE = 0x0412
    ERROR_IGNORED_ALL_SUBSCRIPTIONS = 0x0413
    ERROR_TOO_MANY_SUBSCRIPTIONS = 0x0414
    ERROR_IGNORED_ALL_NOTIFICATIONS = 0x0415
    ERROR_PRINT_SUPPORT_FILE_NOT_FOUND = 0x0416
    ERROR_DOCUMENT_PASSWORD = 0x0417
    ERROR_DOCUMENT_PERMISSION = 0x0418
    ERROR_DOCUMENT_SECURITY = 0x0419
    ERROR_DOCUMENT_UNPRINTABLE = 0x041A
    ERROR_ACCOUNT_INFO_NEEDED = 0x041B
    ERROR_ACCOUNT_CLOSED = 0x041C
    ERROR_ACCOUNT_LIMIT_REACHED = 0x041D
    ERROR_ACCOUNT_AUTHORIZATION_FAILED = 0x041E
    ERROR_NOT_FETCHABLE = 0x041F
    ERROR_CUPS_ACCOUNT_INFO_NEEDED = 0x049C
    ERROR_CUPS_ACCOUNT_CLOSED = 0x049D
    ERROR_CUPS_ACCOUNT_LIMIT_REACHED = 0x049E
    ERROR_CUPS_ACCOUNT_AUTHORIZATION_FAILED = 0x049F
    ERROR_INTERNAL = 0x0500
    ERROR_OPERATION_NOT_SUPPORTED = 0x0501
    ERROR_SERVICE_UNAVAILABLE = 0x0502
    ERROR_VERSION_NOT_SUPPORTED = 0x0503
    ERROR_DEVICE = 0x0504
    ERROR_TEMPORARY = 0x0505
    ERROR_NOT_ACCEPTING_JOBS = 0x0506
    ERROR_BUSY = 0x0507
    ERROR_JOB_CANCELED = 0x0508
    ERROR_MULTIPLE_JOBS_NOT_SUPPORTED = 0x0509
    ERROR_PRINTER_IS_DEACTIVATED = 0x050A
    ERROR_TOO_MANY_JOBS = 0x050B
    ERROR_TOO_MANY_DOCUMENTS = 0x050C
    ERROR_CUPS_AUTHENTICATION_CANCELED = 0x1000
    ERROR_CUPS_PKI = 0x1001
    ERROR_CUPS_UPGRADE_REQUIRED = 0x1002


class IppOperation(IntEnum):
    """Represent the ENUMs of an operation."""

    CUPS_INVALID = -0x0001
    CUPS_NONE = 0x0000
    PRINT_JOB = 0x0002
    PRINT_URI = 0x0003
    VALIDATE_JOB = 0x0004
    CREATE_JOB = 0x0005
    SEND_DOCUMENT = 0x0006
    SEND_URI = 0x0007
    CANCEL_JOB = 0x0008
    GET_JOB_ATTRIBUTES = 0x0009
    GET_JOBS = 0x000A
    GET_PRINTER_ATTRIBUTES = 0x000B
    HOLD_JOB = 0x000C
    RELEASE_JOB = 0x000D
    RESTART_JOB = 0x000E
    PAUSE_PRINTER = 0x0010
    RESUME_PRINTER = 0x0011
    PURGE_JOBS = 0x0012
    SET_PRINTER_ATTRIBUTES = 0x0013
    SET_JOB_ATTRIBUTES = 0x0014
    GET_PRINTER_SUPPORTED_VALUES = 0x0015
    CREATE_PRINTER_SUBSCRIPTIONS = 0x0016
    CREATE_JOB_SUBSCRIPTIONS = 0x0017
    GET_SUBSCRIPTION_ATTRIBUTES = 0x0018
    GET_SUBSCRIPTIONS = 0x0019
    RENEW_SUBSCRIPTION = 0x001A
    CANCEL_SUBSCRIPTION = 0x001B
    GET_NOTIFICATIONS = 0x001C
    SEND_NOTIFICATIONS = 0x001D
    GET_RESOURCE_ATTRIBUTES = 0x001E
    GET_RESOURCE_DATA = 0x001F
    GET_RESOURCES = 0x0020
    GET_PRINT_SUPPORT_FILES = 0x0021
    ENABLE_PRINTER = 0x0022
    DISABLE_PRINTER = 0x0023
    PAUSE_PRINTER_AFTER_CURRENT_JOB = 0x0024
    HOLD_NEW_JOBS = 0x0025
    RELEASE_HELD_NEW_JOBS = 0x0026
    DEACTIVATE_PRINTER = 0x0027
    ACTIVATE_PRINTER = 0x0028
    RESTART_PRINTER = 0x0029
    SHUTDOWN_PRINTER = 0x002A
    STARTUP_PRINTER = 0x002B
    REPROCESS_JOB = 0x002C
    CANCEL_CURRENT_JOB = 0x002D
    SUSPEND_CURRENT_JOB = 0x002E
    RESUME_JOB = 0x002F
    PROMOTE_JOB = 0x0030
    SCHEDULE_JOB_AFTER = 0x0031
    CANCEL_DOCUMENT = 0x0033
    GET_DOCUMENT_ATTRIBUTES = 0x0034
    GET_DOCUMENTS = 0x0035
    DELETE_DOCUMENT = 0x0036
    SET_DOCUMENT_ATTRIBUTES = 0x0037
    CANCEL_JOBS = 0x0038
    CANCEL_MY_JOBS = 0x0039
    RESUBMIT_JOB = 0x003A
    CLOSE_JOB = 0x003B
    IDENTIFY_PRINTER = 0x003C
    VALIDATE_DOCUMENT = 0x003D
    ADD_DOCUMENT_IMAGES = 0x003E
    ACKNOWLEDGE_DOCUMENT = 0x003F
    ACKNOWLEDGE_IDENTIFY_PRINTER = 0x0040
    ACKNOWLEDGE_JOB = 0x0041
    FETCH_DOCUMENT = 0x0042
    FETCH_JOB = 0x0043
    GET_OUTPUT_DEVICE_ATTRIBUTES = 0x0044
    UPDATE_ACTIVE_JOBS = 0x0045
    DEREGISTER_OUTPUT_DEVICE = 0x0046
    UPDATE_DOCUMENT_STATUS = 0x0047
    UPDATE_JOB_STATUS = 0x0048
    UPDATE_OUTPUT_DEVICE_ATTRIBUTES = 0x0049
    GET_NEXT_DOCUMENT_DATA = 0x004A
    ALLOCATE_PRINTER_RESOURCES = 0x004B
    CREATE_PRINTER = 0x004C
    DEALLOCATE_PRINTER_RESOURCES = 0x004D
    DELETE_PRINTER = 0x004E
    GET_PRINTERS = 0x004F
    SHUTDOWN_ONE_PRINTER = 0x0050
    STARTUP_ONE_PRINTER = 0x0051
    CANCEL_RESOURCE = 0x0052
    CREATE_RESOURCE = 0x0053
    INSTALL_RESOURCE = 0x0054
    SEND_RESOURCE_DATA = 0x0055
    SET_RESOURCE_ATTRIBUTES = 0x0056
    CREATE_RESOURCE_SUBSCRIPTIONS = 0x0057
    CREATE_SYSTEM_SUBSCRIPTIONS = 0x0058
    DISABLE_ALL_PRINTERS = 0x0059
    ENABLE_ALL_PRINTERS = 0x005A
    GET_SYSTEM_ATTRIBUTES = 0x005B
    GET_SYSTEM_SUPPORTED_VALUES = 0x005C
    PAUSE_ALL_PRINTERS = 0x005D
    PAUSE_ALL_PRINTERS_AFTER_CURRENT_JOB = 0x005E
    REGISTER_OUTPUT_DEVICE = 0x005F
    RESTART_SYSTEM = 0x0060
    RESUME_ALL_PRINTERS = 0x0061
    SET_SYSTEM_ATTRIBUTES = 0x0062
    SHUTDOWN_ALL_PRINTER = 0x0063
    STARTUP_ALL_PRINTERS = 0x0064
    PRIVATE = 0x4000
    CUPS_GET_DEFAULT = 0x4001
    CUPS_GET_PRINTERS = 0x4002
    CUPS_ADD_MODIFY_PRINTER = 0x4003
    CUPS_DELETE_PRINTER = 0x4004
    CUPS_GET_CLASSES = 0x4005
    CUPS_ADD_MODIFY_CLASS = 0x4006
    CUPS_DELETE_CLASS = 0x4007
    CUPS_ACCEPT_JOBS = 0x4008
    CUPS_REJECT_JOBS = 0x4009
    CUPS_SET_DEFAULT = 0x400A
    CUPS_GET_DEVICES = 0x400B
    CUPS_GET_PPDS = 0x400C
    CUPS_MOVE_JOB = 0x400D
    CUPS_AUTHENTICATE_JOB = 0x400E
    CUPS_GET_PPD = 0x400F
    CUPS_GET_DOCUMENT = 0x4027
    CUPS_CREATE_LOCAL_PRINTER = 0x4028


class IppTag(IntEnum):
    """Represent the ENUMs of a tag."""

    CUPS_INVALID = -1
    ZERO = 0x00
    OPERATION = 0x01
    JOB = 0x02
    END = 0x03
    PRINTER = 0x04
    UNSUPPORTED_GROUP = 0x05
    SUBSCRIPTION = 0x06
    EVENT_NOTIFICATION = 0x07
    RESOURCE = 0x08
    DOCUMENT = 0x09
    SYSTEM = 0x0A
    UNSUPPORTED_VALUE = 0x10
    DEFAULT = 0x11
    UNKNOWN = 0x12
    NO_VALUE = 0x013
    NOT_SETTABLE = 0x15
    DELETE_ATTR = 0x16
    ADMIN_DEFINE = 0x17
    INTEGER = 0x21
    BOOLEAN = 0x22
    ENUM = 0x23
    STRING = 0x30
    DATE = 0x31
    RESOLUTION = 0x32
    RANGE = 0x33
    BEGIN_COLLECTION = 0x34
    TEXT_LANG = 0x35
    NAME_LANG = 0x36
    END_COLLECTION = 0x37
    TEXT = 0x41
    NAME = 0x42
    RESERVED_STRING = 0x43
    KEYWORD = 0x44
    URI = 0x45
    URI_SCHEME = 0x46
    CHARSET = 0x47
    LANGUAGE = 0x48
    MIME_TYPE = 0x49
    MEMBER_NAME = 0x4A
    EXTENSION = 0x7F
    CUPS_MASK = 0x7FFFFFFF
    CUPS_CONST = -0x7FFFFFFF - 1


class IppJobState(IntEnum):
    """Represent the ENUMs of the state of a print job."""

    PENDING = 0x03
    HELD = 0x04
    PROCESSING = 0x05
    STOPPED = 0x06
    CANCELED = 0x07
    ABORTED = 0x08
    COMPLETED = 0x09


class IppDocumentState(IntEnum):
    """Represent the ENUMs of the state of a document."""

    PENDING = 0x03
    PROCESSING = 0x05
    CANCELED = 0x07
    ABORTED = 0x08
    COMPLETED = 0x08


class IppPrinterState(IntEnum):
    """Represent the ENUMs of the state of a printer."""

    IDLE = 0x0003
    PROCESSING = 0x0004
    STOPPED = 0x0005


class IppFinishing(IntEnum):
    """Represent the ENUMs of the finishings attribute."""

    NONE = 0x0003
    STAPLE = 0x0004
    PUNCH = 0x0005
    COVER = 0x0006
    BIND = 0x0007
    SADDLE_STITCH = 0x0008
    EDGE_STITCH = 0x0009
    STAPLE_TOP_LEFT = 0x0014
    STAPLE_BOTTOM_LEFT = 0x0015
    STAPLE_TOP_RIGHT = 0x0016
    STAPLE_BOTTOM_RIGHT = 0x0017
    EDGE_STITCH_LEFT = 0x0018
    EDGE_STITCH_TOP = 0x0019
    EDGE_STITCH_RIGHT = 0x001A
    EDGE_STITCH_BOTTOM = 0x001B
    STAPLE_DUAL_LEFT = 0x001C
    STAPLE_DUAL_TOP = 0x001D
    STAPLE_DUAL_RIGHT = 0x001E
    STAPLE_DUAL_BOTTOM = 0x001F


class IppPrintQuality(IntEnum):
    """Represent the ENUMs of the print-quality attribute."""

    DRAFT = 0x0003
    NORMAL = 0x0004
    HIGH = 0x0005


class IppOrientationRequested(IntEnum):
    """Represent the ENUMs of the orientation-requested attribute."""

    PORTRAIT = 0x0003
    LANDSCAPE = 0x0004
    REVERSE_LANDSCAPE = 0x0005
    REVERSE_PORTRAIT = 0x0006


ATTRIBUTE_ENUM_MAP = {
    "document-state": IppDocumentState, # PWG5100.5
    "finishings": IppFinishing, # RFC8011
    "finishings-default": IppFinishing, # RFC8011
    "finishings-supported": IppFinishing, # RFC8011
    "job-state": IppJobState, # RFC8011
    "media-source-feed-orientation": IppOrientationRequested, # PWG5100.7
    "operations-supported": IppOperation, # RFC8011
    "orientation-requested": IppOrientationRequested, # RFC8011
    "orientation-requested-default": IppOrientationRequested, # RFC8011
    "orientation-requested-supported": IppOrientationRequested, # RFC8011
    "printer-state": IppPrinterState, # RFC8011
    "print-quality": IppPrintQuality, # RFC8011
    "print-quality-default": IppPrintQuality, # RFC8011
    "print-quality-supported": IppPrintQuality, # RFC8011
    "status-code": IppStatus, # RFC8011
}



"""Attribute Tags for IPP."""
from .enums import IppTag

ATTRIBUTE_TAG_MAP = {
    "attributes-charset": IppTag.CHARSET,
    "attributes-natural-language": IppTag.LANGUAGE,
    "document-number": IppTag.INTEGER,
    "printer-uri": IppTag.URI,
    "requesting-user-name": IppTag.NAME,
    "job-id": IppTag.INTEGER,
    "document-name": IppTag.NAME,
    "job-name": IppTag.NAME,
    "document-format": IppTag.MIME_TYPE,
    "last-document": IppTag.BOOLEAN,
    "copies": IppTag.INTEGER,
    "job-cancel-after": IppTag.INTEGER,
    "job-hold-until": IppTag.KEYWORD,
    "job-k-octets": IppTag.INTEGER,
    "job-impressions-completed": IppTag.INTEGER,
    "job-media-sheets-completed": IppTag.INTEGER,
    "job-originating-host-name": IppTag.NAME,
    "job-originating-user-name": IppTag.NAME,
    "job-printer-state-message": IppTag.TEXT,
    "job-printer-state-reasons": IppTag.KEYWORD,
    "job-priority": IppTag.INTEGER,
    "number-up": IppTag.INTEGER,
    "job-sheets": IppTag.NAME,
    "job-uri": IppTag.URI,
    "job-state": IppTag.ENUM,
    "job-state-reasons": IppTag.KEYWORD,
    "job-uuid": IppTag.URI,
    "requested-attributes": IppTag.KEYWORD,
    "member-uris": IppTag.URI,
    "operations-supported": IppTag.ENUM,
    "ppd-name": IppTag.NAME,
    "printer-state-reason": IppTag.KEYWORD,
    "printer-is-shared": IppTag.BOOLEAN,
    "printer-error-policy": IppTag.NAME,
    "printer-geo-location": IppTag.URI,
    "printer-info": IppTag.TEXT,
    "printer-organization": IppTag.TEXT,
    "printer-organizational-unit": IppTag.TEXT,
    "which-jobs": IppTag.KEYWORD,
    "my-jobs": IppTag.BOOLEAN,
    "purge-jobs": IppTag.BOOLEAN,
    "hold-job-until": IppTag.KEYWORD,
    "job-printer-uri": IppTag.URI,
    "printer-location": IppTag.TEXT,
    "printer-state": IppTag.ENUM,
    "printer-state-reasons": IppTag.KEYWORD,
    "printer-up-time": IppTag.INTEGER,
    "printer-uri-supported": IppTag.URI,
    "document-state": IppTag.ENUM,
    "device-uri": IppTag.URI,
    "ipp-attribute-fidelity": IppTag.BOOLEAN,
    "finishings": IppTag.ENUM,
    "orientation-requested": IppTag.ENUM,
    "print-quality": IppTag.ENUM,
    "time-at-creation": IppTag.INTEGER,
    "time-at-processing": IppTag.INTEGER,
    "time-at-completed": IppTag.INTEGER,
    "media": IppTag.NAME,
    "center-of-pixel": IppTag.BOOLEAN,
}



"""Tests for IPP Models."""
# pylint: disable=R0912,R0915
from __future__ import annotations

from typing import Any, List

import pytest

from pyipp import models, parser

from . import IPPE10_PRINTER_ATTRS, load_fixture_binary


@pytest.mark.asyncio
async def test_info() -> None:  # noqa: PLR0915
    """Test Info model."""
    parsed = parser.parse(load_fixture_binary("get-printer-attributes-epsonxp6000.bin"))
    data = parsed["printers"][0]
    info = models.Info.from_dict(data)

    assert info
    assert info.command_set == "ESCPL2,BDC,D4,D4PX,ESCPR7,END4,GENEP,URF"
    assert info.location == ""
    assert info.name == "EPSON XP-6000 Series"
    assert info.manufacturer == "EPSON"
    assert info.model == "XP-6000 Series"
    assert info.printer_name == "ipp/print"
    assert info.printer_info == "EPSON XP-6000 Series"
    assert info.printer_uri_supported == [
        "ipps://192.168.1.92:631/ipp/print",
        "ipp://192.168.1.92:631/ipp/print",
    ]
    assert info.more_info == "http://192.168.1.92:80/PRESENTATION/BONJOUR"
    assert info.serial == "583434593035343012"
    assert info.uuid == "cfe92100-67c4-11d4-a45f-f8d027761251"
    assert info.version == "20.44.NU25M7"
    assert info.uptime == 783801

    # no make/model, device id
    data["printer-make-and-model"] = ""
    info = models.Info.from_dict(data)

    assert info
    assert info.name == "EPSON XP-6000 Series"
    assert info.printer_name == "ipp/print"
    assert info.manufacturer == "EPSON"
    assert info.model == "XP-6000 Series"

    # no make/model, no device id, URI name
    data["printer-device-id"] = ""
    data["printer-make-and-model"] = ""
    data["printer-name"] = "ipp/print"
    info = models.Info.from_dict(data)

    assert info
    assert info.name == "IPP Printer"
    assert info.printer_name == "ipp/print"
    assert info.manufacturer == "Unknown"
    assert info.model == "Unknown"

    # no make/model, no device id, name
    data["printer-device-id"] = ""
    data["printer-make-and-model"] = ""
    data["printer-name"] = "Printy"
    info = models.Info.from_dict(data)

    assert info
    assert info.name == "Printy"
    assert info.printer_name == "Printy"
    assert info.manufacturer == "Unknown"
    assert info.model == "Unknown"

    # no make/model, no device id, no name
    data["printer-device-id"] = ""
    data["printer-make-and-model"] = ""
    data["printer-name"] = ""
    info = models.Info.from_dict(data)

    assert info
    assert info.name == "IPP Printer"
    assert info.printer_name == ""
    assert info.manufacturer == "Unknown"
    assert info.model == "Unknown"


@pytest.mark.asyncio
async def test_state() -> None:
    """Test State model."""
    data: dict[str, Any] = {
        "printer-state": 4,
        "printer-state-reasons": "none",
    }

    state = models.State.from_dict(data)

    assert state
    assert state.printer_state == "printing"
    assert state.reasons is None
    assert state.message is None


@pytest.mark.asyncio
async def test_printer() -> None:  # noqa: PLR0915
    """Test Printer model."""
    parsed = parser.parse(load_fixture_binary("get-printer-attributes-epsonxp6000.bin"))
    printer = models.Printer.from_dict(parsed["printers"][0])

    assert printer

    assert printer.info
    assert printer.info.command_set == "ESCPL2,BDC,D4,D4PX,ESCPR7,END4,GENEP,URF"
    assert printer.info.location == ""
    assert printer.info.name == "EPSON XP-6000 Series"
    assert printer.info.manufacturer == "EPSON"
    assert printer.info.model == "XP-6000 Series"
    assert printer.info.printer_name == "ipp/print"
    assert printer.info.printer_info == "EPSON XP-6000 Series"
    assert printer.info.printer_uri_supported == [
        "ipps://192.168.1.92:631/ipp/print",
        "ipp://192.168.1.92:631/ipp/print",
    ]
    assert printer.info.more_info == "http://192.168.1.92:80/PRESENTATION/BONJOUR"
    assert printer.info.serial == "583434593035343012"
    assert printer.info.uuid == "cfe92100-67c4-11d4-a45f-f8d027761251"
    assert printer.info.version == "20.44.NU25M7"
    assert printer.info.uptime == 783801

    assert printer.state
    assert printer.state.printer_state == "idle"
    assert printer.state.reasons == "marker-supply-low-warning"
    assert printer.state.message is None

    assert printer.markers
    assert isinstance(printer.markers, list)
    assert len(printer.markers) == 5

    assert printer.markers[0]
    assert printer.markers[0].marker_id == 4
    assert printer.markers[0].marker_type == "ink-cartridge"
    assert printer.markers[0].name == "Black ink"
    assert printer.markers[0].color == "#000000"
    assert printer.markers[0].level == 64
    assert printer.markers[0].low_level == 15
    assert printer.markers[0].high_level == 100

    assert printer.markers[1]
    assert printer.markers[1].marker_id == 1
    assert printer.markers[1].marker_type == "ink-cartridge"
    assert printer.markers[1].name == "Cyan ink"
    assert printer.markers[1].color == "#00FFFF"
    assert printer.markers[1].level == 99
    assert printer.markers[1].low_level == 15
    assert printer.markers[1].high_level == 100

    assert printer.markers[2]
    assert printer.markers[2].marker_id == 2
    assert printer.markers[2].marker_type == "ink-cartridge"
    assert printer.markers[2].name == "Magenta ink"
    assert printer.markers[2].color == "#FF00FF"
    assert printer.markers[2].level == 83
    assert printer.markers[2].low_level == 15
    assert printer.markers[2].high_level == 100

    assert printer.markers[3]
    assert printer.markers[3].marker_id == 0
    assert printer.markers[3].marker_type == "ink-cartridge"
    assert printer.markers[3].name == "Photo Black ink"
    assert printer.markers[3].color == "#000000"
    assert printer.markers[3].level == 27
    assert printer.markers[3].low_level == 15
    assert printer.markers[3].high_level == 100

    assert printer.markers[4]
    assert printer.markers[4].marker_id == 3
    assert printer.markers[4].marker_type == "ink-cartridge"
    assert printer.markers[4].name == "Yellow ink"
    assert printer.markers[4].color == "#FFFF00"
    assert printer.markers[4].level == 6
    assert printer.markers[4].low_level == 15
    assert printer.markers[4].high_level == 100

    assert printer.uris
    assert isinstance(printer.uris, list)
    assert len(printer.uris) == 2

    assert printer.uris[0]
    assert printer.uris[0].uri == "ipps://192.168.1.92:631/ipp/print"
    assert printer.uris[0].authentication is None
    assert printer.uris[0].security == "tls"

    assert printer.uris[1]
    assert printer.uris[1].uri == "ipp://192.168.1.92:631/ipp/print"
    assert printer.uris[1].authentication is None
    assert printer.uris[1].security is None

def test_printer_as_dict() -> None:
    """Test the dictionary version of Printer."""
    parsed = parser.parse(load_fixture_binary("get-printer-attributes-epsonxp6000.bin"))
    printer = models.Printer.from_dict(parsed["printers"][0])

    assert printer

    printer_dict = printer.as_dict()
    assert printer_dict
    assert isinstance(printer_dict, dict)
    assert isinstance(printer_dict["info"], dict)
    assert isinstance(printer_dict["state"], dict)
    assert isinstance(printer_dict["markers"], List)
    assert len(printer_dict["markers"]) == 5
    assert isinstance(printer_dict["uris"], List)
    assert len(printer_dict["uris"]) == 2

def test_printer_update_from_dict() -> None:
    """Test updating data of Printer."""
    parsed = parser.parse(load_fixture_binary("get-printer-attributes-epsonxp6000.bin"))
    printer = models.Printer.from_dict(parsed["printers"][0])

    assert printer
    assert printer.info
    assert printer.info.uptime == 783801

    parsed["printer-up-time"] = 2
    printer.update_from_dict(parsed)

    assert printer
    assert printer.info
    assert printer.info.uptime == 2

@pytest.mark.asyncio
async def test_printer_with_single_marker() -> None:
    """Test Printer model with single marker."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["marker-names"] = "Black"
    data["marker-types"] = "ink-cartridge"
    data["marker-colors"] = "#FF0000"
    data["marker-levels"] = 77
    data["marker-high-levels"] = 100
    data["marker-low-levels"] = 0

    printer = models.Printer.from_dict(data)
    assert printer
    assert printer.markers[0]
    assert printer.markers[0].name == "Black"
    assert printer.markers[0].color == "#FF0000"
    assert printer.markers[0].level == 77
    assert printer.markers[0].high_level == 100
    assert printer.markers[0].low_level == 0
    assert printer.markers[0].marker_type == "ink-cartridge"


@pytest.mark.asyncio
async def test_printer_with_single_marker_empty_strings() -> None:
    """Test Printer model with single marker with empty string values."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["marker-names"] = ""
    data["marker-types"] = ""
    data["marker-colors"] = ""
    data["marker-levels"] = ""
    data["marker-low-levels"] = ""
    data["marker-high-levels"] = ""

    printer = models.Printer.from_dict(data)
    assert printer
    assert len(printer.markers) == 0


@pytest.mark.asyncio
async def test_printer_with_single_marker_invalid() -> None:
    """Test Printer model with single invalid marker name."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["marker-names"] = -1

    printer = models.Printer.from_dict(data)
    assert printer
    assert len(printer.markers) == 0


@pytest.mark.asyncio
async def test_printer_with_extra_marker_data() -> None:
    """Test Printer model with extra marker data."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["marker-names"] = ["Black"]
    data["marker-types"] = ["ink-cartridge", "ink"]
    data["marker-colors"] = ["#FF0000", "#FF1111"]
    data["marker-levels"] = [99, 33]
    data["marker-low-levels"] = [0, 10]
    data["marker-high-levels"] = [99, 100]

    printer = models.Printer.from_dict(data)
    assert printer
    assert len(printer.markers) == 1
    assert printer.markers[0]
    assert printer.markers[0].name == "Black"
    assert printer.markers[0].color == "#FF0000"
    assert printer.markers[0].level == 99
    assert printer.markers[0].high_level == 99
    assert printer.markers[0].low_level == 0
    assert printer.markers[0].marker_type == "ink-cartridge"


@pytest.mark.asyncio
async def test_printer_with_single_supported_uri() -> None:
    """Test Printer model with single supported uri."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["printer-uri-supported"] = "ipp://10.104.12.95:631/ipp/print"
    data["uri-authentication-supported"] = "none"
    data["uri-security-supported"] = "none"

    printer = models.Printer.from_dict(data)
    assert printer
    assert printer.uris[0]
    assert printer.uris[0].uri == "ipp://10.104.12.95:631/ipp/print"
    assert printer.uris[0].authentication is None
    assert printer.uris[0].security is None


@pytest.mark.asyncio
async def test_printer_with_single_supported_uri_extra_data() -> None:
    """Test Printer model with single supported uri with extra data."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["printer-uri-supported"] = "ipp://10.104.12.95:631/ipp/print"
    data["uri-authentication-supported"] = ["none", "basic"]
    data["uri-security-supported"] = ["none", "tls"]

    printer = models.Printer.from_dict(data)
    assert printer
    assert printer.uris[0]
    assert printer.uris[0].uri == "ipp://10.104.12.95:631/ipp/print"
    assert printer.uris[0].authentication is None
    assert printer.uris[0].security is None


@pytest.mark.asyncio
async def test_printer_with_single_supported_uri_invalid_uri() -> None:
    """Test Printer model with single invalid supported uri."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["printer-uri-supported"] = -1
    data["uri-authentication-supported"] = "none"
    data["uri-security-supported"] = "none"

    printer = models.Printer.from_dict(data)
    assert printer
    assert len(printer.uris) == 0


@pytest.mark.asyncio
async def test_printer_with_single_supported_uri_with_security() -> None:
    """Test Printer model with multiple markers."""
    data = IPPE10_PRINTER_ATTRS.copy()
    data["printer-uri-supported"] = "ipps://10.104.12.95:631/ipp/print"
    data["uri-authentication-supported"] = "basic"
    data["uri-security-supported"] = "tls"

    printer = models.Printer.from_dict(data)
    assert printer
    assert printer.uris[0]
    assert printer.uris[0].uri == "ipps://10.104.12.95:631/ipp/print"
    assert printer.uris[0].authentication == "basic"
    assert printer.uris[0].security == "tls"
